package ca.virology.japafi.algorithms;

import java.util.*;

import java.io.*;

import ca.virology.japafi.structures.*;


/* This class implements the APBT algorithm for finding common approximate patterns
 * in a following way: given the first string, 
 * where potential start positions of a pattern are marked in a supplied BitSet,
 *   it determines the positions in the FIRST STRING, 
 *   where substrings of length minLength match at least one substring 
 *   in the second string with up to maxDiff differences.
 * The output of the algorithm is a new BitSet with positions 
 * in the FIRST string where pattern substrings start 
 * 
 * Max length of patterns is set to 300 characters. 
 * 
 *  @author marina
*/

public class APBTBoolean
{	
	/** This constant defines the size of a chunk of the matching matrix to be processed simultaneously. 
	 * This speeds up the calculation by decreasing the memory used. 
	 * The simultaneously processed part of a matrix is of size: 
	 * CHUNK_SIZE*_length1, where _length1 is the length of the first string.
	 */
	public final int CHUNK_SIZE=1000;
	
	/** This variable defines the maximum length of matches. 
	 */
	public int MAX_LENGTH=300;
	
	int _chunkSize;
	int _maxLength;
	
	BitSet _oldStartPositions;	
	BitSet _newStartPositions;
	
	int _maxDiff;
	int _minLength;
	
	Map _charPositions2;

	boolean [][] _matrix;
	
	int _length1;
	int _length2;
	char [] _seq1;
	char [] _seq2;
	
	Object[][]  _state;		
	
	int _currentLength2; //including max length
	/** Constructor with char arrays instead of strings - to avoid the copying of big strings 
	 * when processing multiple times.
	 * @param seq1arr - char array representing string 1
	 * @param seq2arr - char array representing string 2
	 * @param minLength - the minimum length of a solution pattern
	 * @param maxDiff - the maximum number of allowed errors
	 */
	public APBTBoolean(BitSet oldStartPositions,char [] seq1arr, char [] seq2arr, 
			int minLength, int maxDiff)
	{		
		_oldStartPositions=oldStartPositions;
		_newStartPositions=new BitSet(seq1arr.length);
		_seq1=seq1arr;
		_seq2=seq2arr;
		_length1=_seq1.length;
		_length2=_seq2.length;
		_matrix=new boolean[_length1][];
		_minLength=minLength;
		_maxDiff=maxDiff;
		_maxLength=MAX_LENGTH;
		_chunkSize=CHUNK_SIZE;
	}
	
	private boolean betterPathExists(int currlen, int currdiff,int currI, int currJ, int i,int j)
	{
		if(currlen>1)
		{						
			int []  instate=(int [])_state[i%_maxLength][j];			
			
			if(instate==null)
			{				
				instate=new int [2*_maxDiff+1];	
				int deltaI=i-currI;
				int deltaJ=j-currJ;
				int diagonalID=deltaI-deltaJ;
				instate[_maxDiff+diagonalID]=currdiff+1;
				_state[i%_maxLength][j]=instate;
				return false;
				
			}
			else //check what is the best (minimum diff) checked for this point
			{					
				int deltaI=i-currI;
				int deltaJ=j-currJ;
				int diagonalID=deltaI-deltaJ;
				
				if(instate[_maxDiff+diagonalID]==0 ||
						instate[_maxDiff+diagonalID]>currdiff+1 )
				{
					instate[_maxDiff+diagonalID]=currdiff+1;
					_state[i%_maxLength][j]=instate;
					return false;
				}
				else
					return true;								
				
			}				
		
				
		}
		return false;
	}
	
	
	/**
	* The main recursion - expanding of the path. 
	* Stop conditions: the path cannot be expanded without an additional error,
	* OR a better path has been collected ending at point (currI, currJ)
	*/

	private void continuePath(int startI, int startJ, int currI, int currJ, int currlen, int currdiff,int shiftJ,boolean continueFurther)
	{		
		if(currlen>=_maxLength)
			return;
		

		if(continueFurther)		
		{			
			
			if(betterPathExists(currlen, currdiff,startI, startJ,currI,currJ))				
				return;
			//define bounds of the target square
			int LT_I=currI+1;
			if(LT_I==_length1)
				return;
			int LT_J=currJ+1;
			if(LT_J==_currentLength2)
				return;;
	
			int RT_J=Math.min(LT_J+_maxDiff+1-currdiff,_currentLength2);
			int LB_I=Math.min(LT_I+_maxDiff+1-currdiff,_length1);
		
			int RB_J=RT_J;		
	
			int currJBound=RB_J;
			int currIBound=LB_I;	 
			
			//explore the main diagonal
			boolean stop=false;
	
			for(int i=currI+1,j=currJ+1;
					i<LB_I && 
					j<RT_J && !stop;i++,j++)
			{
				if(_matrix[i][j])
				{					
					continuePath(startI,startJ,i,j,Math.min(i-startI,j-startJ)+1, Math.max(i-currI,j-currJ)-1+currdiff,shiftJ,true);
					stop=true;
					currJBound=j;
					currIBound=i;
				}
			}
				
			for(int i=currIBound+1;i<LB_I;i++)
			{
				if(_matrix[i][currJBound])			
					continuePath(startI,startJ,i,currJBound,Math.min(i-startI,currJBound-startJ)+1, Math.max(i-currI,currJBound-currJ)-1+currdiff,shiftJ,false);
			}
			
			for(int j=currJBound+1;j<RT_J;j++)
			{
				if(_matrix[currIBound][j])						
					continuePath(startI,startJ,currIBound,j,Math.min(currIBound-startI,j-startJ)+1, Math.max(currIBound-currI,j-currJ)-1+currdiff,shiftJ,false);
			}
	
			//add an edge from next k upper and lower diagonals		
			for(int k=1;k<_maxDiff+1-currdiff ;k++)
			{
				//lower diagonals			
				stop=false;
				for(int i=LT_I+k, j=LT_J;i<LB_I && j<currJBound && !stop;i++,j++)
				{
					if(_matrix[i][j])
					{
									
						continuePath(startI,startJ,i,j,Math.min(i-startI,j-startJ)+1, Math.max(i-currI,j-currJ)-1+currdiff,shiftJ,true);
						if(j<currJBound)
						{
							currJBound=j;
							for(int m=i+1;m<LB_I;m++)
							{
								if(_matrix[m][currJBound])
									continuePath(startI,startJ,m,currJBound,Math.min(m-startI,currJBound-startJ)+1, Math.max(m-currI,currJBound-currJ)-1+currdiff,shiftJ,false);
							}
							
						}
						stop=true;
					}
				}
	
				//upper diagonals	
				stop=false;
				for(int i=LT_I, j=LT_J+k;i<currIBound && j<RT_J && !stop;i++,j++)
				{
					if(_matrix[i][j])
					{
						continuePath(startI,startJ,i,j,Math.min(i-startI,j-startJ)+1, Math.max(i-currI,j-currJ)-1+currdiff,shiftJ,true);
						if(i<currIBound)
						{
							currIBound=i;
							for(int m=j+1;m<RT_J;m++)
							{
								
								if(_matrix[currIBound][m])
									continuePath(startI,startJ,currIBound,m,Math.min(currIBound-startI,m-startJ)+1, Math.max(currIBound-currI,m-currJ)-1+currdiff,shiftJ,false);
							}
							
						}
							
						stop=true;
					}
				}
			}	
		}
		
//		the maximal path can not continue. Check if this is a solution
		if(currlen>=_minLength )
		{	
			markNewStartPositions(startI,  startJ+shiftJ,  currI,  currJ+shiftJ);
		}

	}	

	
	


	/********************************************************************************************
	 *  PUBLIC 
	 ********************************************************************************************/

	
	/**
	 * This function processes the matching matrix for specified start positions
	 * in a row major order.
	 * The processing is performed by chunks.
	 *
	 */	
	public void process()
	{			
		int fullNumberOfChunks=_length2/_chunkSize;
		int remaining=_length2%_chunkSize;
		int k=0;
		for(;k<fullNumberOfChunks;k++)
		{
			int startJ=k*_chunkSize;
			process(startJ);			
		}
		if(remaining>0)
		{
			int startJ=k*_chunkSize;
			process(startJ);
		}
	}
	
	/**
	 * This function processes the matching matrix for specified start positions
	 * in a column major order (by reversing 2 strings). 
	 * The processing is performed by chunks.
	 *
	 */	
	
	
	/**
	 * 
	 * @return BitSet of start positions of patterns in the FIRST string.
	 */
	public BitSet getNewStartPositions()
	{
		
		return _newStartPositions;
	}
	
	
	
	/********************************************************************************************
	 *  PRIVATE
	 ********************************************************************************************/
	
	/** This function processes the parts of the matrix starting from particular position in the second string.
	* @param startJ - start position in a second string
	*/
	
	private void process(int startJ)
	{
	
		if(startJ>_length2-_minLength)
			return;
		int toToProcess=Math.min(startJ+_chunkSize, _length2-_minLength+1);
		
		int toToExpand=Math.min(startJ+_chunkSize+_maxLength+1, _length2);
		_currentLength2=toToExpand-startJ;
		
		initializeMatrix(startJ,toToExpand);
		
		for(int i=0;i<=_length1-_minLength;i++)//i-row
		{	
			if(_oldStartPositions.get(i) && !_newStartPositions.get(i))
			{
				for(int j=0;j<(toToProcess-startJ);j++)
				{				
					
					if(_matrix[i][j])
					{					
						createPaths(i,j,startJ);				
					}
				}
			//reset the row of _state array to use with the next added (i+_maxLength)-th row: 
			}
			_state[(i)%_maxLength]=new Object[_currentLength2];
			
		}		
		
	}
	
	
	/** Initialises a new path of ML=1, EN=0, starting from current true cell of the matrix
	*/	
	private void createPaths(int startI, int startJ ,int shiftJ)
	{
		
		continuePath(startI,startJ,startI,startJ,1,0,shiftJ,true);
	}		
	
	private void initializeMatrix(int from, int to)
	{
		_charPositions2=new HashMap(20);
		for(int j=from;j<to;j++)
		{
			char curr=_seq2[j];
			boolean [] row=(boolean [])_charPositions2.get(new Character(curr));
			if(row==null)
				row=new boolean[to-from];
			row[j-from]=true;
			_charPositions2.put(new Character(curr),row);
		}		

		for(int i=0;i<_length1;i++)
		{
			char curr=_seq1[i];
			if(_charPositions2.get(new Character(curr))!=null)
			{
				boolean [] row=(boolean [])_charPositions2.get(new Character(curr));
				_matrix[i]=row;
			}
			else
				_matrix[i]=new boolean[to-from];
		}
		
		_state=new Object[_maxLength][_currentLength2];
			
	}	
	
	private void markNewStartPositions(int startI, int startJ, int currI, int currJ)
	{		
		int lastStart=currI-_minLength+1;
		for(int i=startI;i<=lastStart;i++)
		{
			if(_oldStartPositions.get(i))
				_newStartPositions.set(i);
		}		
	}
	


}





