package ca.virology.japafi.algorithms;


import java.util.*;

import java.io.*;

import ca.virology.japafi.structures.*;


/* This class implements the APBT algorithm for answering the question
 * whether a given pattern can match at least one substring 
 * of length at least minLength 
 * in a given text with up to maxDiff differences.
 * 
 * The output of the algorithm is an answer: 
 * true if pattern can be matched, false otherwise 
 * 
 * Max length of patterns is set to 300 characters. 
 * 
 *  @author marina
*/

public class APBTPatternBoolean
{	
	/** This constant defines the size of a chunk of the matching matrix to be processed simultaneously. 
	 * This speeds up the calculation by decreasing the memory used. 
	 * The simultaneously processed part of a matrix is of size: 
	 * CHUNK_SIZE*_length1, where _length1 is the length of the first string.
	 *//*
	public final int CHUNK_SIZE=1000;*/
	
	/** This variable defines the maximum length of matches. 
	 */
	public int MAX_LENGTH=300;
	
	//int _chunkSize;
	int _maxLength;	
	
	
	int _maxDiff;
	int _minLength;
	
	Map _charPositions2;

	boolean [][] _matrix;
	
	int _textLength;
	int _patternLength;
	char [] _pattern;
	char [] _text;
	
	Object[][]  _state;		
	
	private boolean _matchFound;
	/** Constructor with char arrays instead of strings - to avoid the copying of big strings 
	 * when processing multiple times.
	 * @param textarr - char array representing string 1
	 * @param patternarr - char array representing string 2
	 * @param minLength - the minimum length of a solution pattern
	 * @param maxDiff - the maximum number of allowed errors
	 */
	public APBTPatternBoolean(char [] textarr, char [] patternarr, 
			int minLength, int maxDiff)
	{		
		_pattern=patternarr;
		_text=textarr;
		_textLength=_text.length;
		_patternLength=_pattern.length;
		_matrix=new boolean[_textLength][];
		_minLength=minLength;
		_maxDiff=maxDiff;
		_maxLength=MAX_LENGTH;
		
	}
	
	private boolean betterPathExists(int currlen, int currdiff,int currI,  int i,int j)
	{
		if(currlen>1)
		{						
			int []  instate=(int [])_state[i%_maxLength][j];			
			
			if(instate==null)
			{				
				instate=new int [2*_maxDiff+1];	
				int deltaI=i-currI;
				int deltaJ=j-0;
				int diagonalID=deltaI-deltaJ;
				instate[_maxDiff+diagonalID]=currdiff+1;
				_state[i%_maxLength][j]=instate;
				return false;
				
			}
			else //check what is the best (minimum diff) checked for this point
			{					
				int deltaI=i-currI;
				int deltaJ=j-0;
				int diagonalID=deltaI-deltaJ;
				
				if(instate[_maxDiff+diagonalID]==0 ||
						instate[_maxDiff+diagonalID]>currdiff+1 )
				{
					instate[_maxDiff+diagonalID]=currdiff+1;
					_state[i%_maxLength][j]=instate;
					return false;
				}
				else
					return true;								
				
			}				
		
				
		}
		return false;
	}
	
	
	/**
	* The main recursion - expanding of the path starting at startI. 
	* Stop conditions: the path cannot be expanded without an additional error,
	* OR a better path has been collected ending at point (currI, currJ)
	*/

	private void continuePath(int startI, int currI, int currJ, int currlen, int currdiff,boolean continueFurther)
	{		
		if(currlen>=_minLength)
		{
			_matchFound=true;
			return ;
		}
		

		if(continueFurther)		
		{			
			if(betterPathExists(currlen, currdiff,startI, currI,currJ))
				return;
							
			{
				//define bounds of the target square
				int LT_I=currI+1;
				if(LT_I==_textLength)
					return;
				int LT_J=currJ+1;
				if(LT_J==_patternLength)
					return;;
		
				int RT_J=Math.min(LT_J+_maxDiff+1-currdiff,_patternLength);
				int LB_I=Math.min(LT_I+_maxDiff+1-currdiff,_textLength);
			
				int RB_J=RT_J;		
		
				int currJBound=RB_J;
				int currIBound=LB_I;	 
				
				//explore the main diagonal
				boolean stop=false;
		
				for(int i=currI+1,j=currJ+1;
						i<Math.min(currI+_maxDiff+2-currdiff,_textLength) && 
						j<Math.min(currJ+_maxDiff+2-currdiff,_patternLength) && !stop;i++,j++)
				{
					if(_matrix[i][j])
					{					
						continuePath(startI,i,j,Math.min(i-startI,j-0)+1, Math.max(i-currI,j-currJ)-1+currdiff,true);
						stop=true;
						currJBound=j;
						currIBound=i;
					}
				}
					
				for(int i=currIBound+1;i<LB_I;i++)
				{
					if(_matrix[i][currJBound])			
						continuePath(startI,i,currJBound,Math.min(i-startI,currJBound-0)+1, Math.max(i-currI,currJBound-currJ)-1+currdiff,false);
				}
				
				for(int j=currJBound+1;j<RT_J;j++)
				{
					if(_matrix[currIBound][j])						
						continuePath(startI,currIBound,j,Math.min(currIBound-startI,j-0)+1, Math.max(currIBound-currI,j-currJ)-1+currdiff,false);
				}
		
				//add an edge from next k upper and lower diagonals		
				for(int k=1;k<_maxDiff+1-currdiff ;k++)
				{
					//lower diagonals			
					stop=false;
					for(int i=LT_I+k, j=LT_J;i<LB_I && j<currJBound && !stop;i++,j++)
					{
						if(_matrix[i][j])
						{
										
							continuePath(startI,i,j,Math.min(i-startI,j-0)+1, Math.max(i-currI,j-currJ)-1+currdiff,true);
							if(j<currJBound)
							{
								currJBound=j;
								for(int m=i+1;m<LB_I;m++)
								{
									if(_matrix[m][currJBound])
										continuePath(startI,m,currJBound,Math.min(m-startI,currJBound-0)+1, Math.max(m-currI,currJBound-currJ)-1+currdiff,false);
								}
								
							}
							stop=true;
						}
					}
		
					//upper diagonals	
					stop=false;
					for(int i=LT_I, j=LT_J+k;i<currIBound && j<RT_J && !stop;i++,j++)
					{
					
						if(_matrix[i][j])
						{
							continuePath(startI,i,j,Math.min(i-startI,j-0)+1, Math.max(i-currI,j-currJ)-1+currdiff,true);
							if(i<currIBound)
							{
								currIBound=i;
								for(int m=j+1;m<RT_J;m++)
								{
									
									if(_matrix[currIBound][m])
										continuePath(startI,currIBound,m,Math.min(currIBound-startI,m-0)+1, Math.max(currIBound-currI,m-currJ)-1+currdiff,false);
								}
								
							}
								
							stop=true;
						}
					}
				}
			}
		}
		
//		the maximal path can not continue. Check if this is a solution
		if(currlen>=_minLength )
		{	
			_matchFound=true;
			
		}	
		

	}	

	
	


	/********************************************************************************************
	 *  PUBLIC 
	 ********************************************************************************************/

	
	/**
	 * This function processes the matching matrix of pattern and text
	 * in a row major order.
	 * Pattern is the horizontal dimension of the matrix, text is a vertical one. 
	 *
	 */	
	public void process()
	{		
		initializeMatrix();
		
		for(int i=0;i<=_textLength-_minLength;i++)
		{			
			if(_matrix[i][0])
			{					
				createPaths(i);
				if(_matchFound)
					return;				
			}
			//reset the row of _state array to use with the next added (i+_maxLength)-th row: 
			
			_state[(i)%_maxLength]=new Object[_patternLength];
			
		}		
	}	
	
	public boolean matchFound()
	{
		return _matchFound;
	}
	
	
	/********************************************************************************************
	 *  PRIVATE
	 ********************************************************************************************/
	
	
	
	/** Initializes a new path of ML=1, EN=0, starting from current true cell of the matrix
	*/	
	private void createPaths(int startI)
	{
		
		continuePath(startI,startI,0,1,0,true);
	}		
	
	private void initializeMatrix()
	{
		_charPositions2=new HashMap(20);
		for(int j=0;j<_patternLength;j++)
		{
			char curr=_pattern[j];
			boolean [] row=(boolean [])_charPositions2.get(new Character(curr));
			if(row==null)
				row=new boolean[_patternLength];
			row[j]=true;
			_charPositions2.put(new Character(curr),row);
		}		

		for(int i=0;i<_textLength;i++)
		{
			char curr=_text[i];
			if(_charPositions2.get(new Character(curr))!=null)
			{
				boolean [] row=(boolean [])_charPositions2.get(new Character(curr));
				_matrix[i]=row;
			}
			else
				_matrix[i]=new boolean[_patternLength];
		}
		
		_state=new Object[_maxLength][_patternLength];
		
	}	
	
	
	


}





