package ca.virology.japafi.algorithms;
import java.util.*;
import ca.virology.japafi.utils.*;
/**
 * This calss performs exact match fitering due to Myers algorithm for a set of
 * input sequences. It finds all substrings in the FIRST sequence 
 * which may potentially match at least one substring of length minLen 
 * in every input sequence with up to maxDiff errors. The oldStartPositions supplies 
 * previous start positions obtained by preprocessing or specified by user
 * 
 * The output of an algorithm is the BitSet where the potential
 * start positions in the first input sequence are marked as true;
 * @author marina
 *
 */
public class Filter 
{
	private static final int MIN_WORD_LENGTH=9;
	
	private BitSet [] _startPositions=new BitSet[2];
	private int current=1;
	private int previous=0;
	private int _minLength;
	private int _maxDiff;
	private char [][] _sequences;
	private HashMap [] _wordMaps;
	private int _wordLength;
	
	public Filter(BitSet oldStartPositions,char [][] sequences, int minLen, int maxDiff)
	{
		_startPositions[previous]=oldStartPositions;
		_startPositions[current]=new BitSet(sequences[0].length);
		_sequences=sequences;
		_minLength=minLen;
		_maxDiff=maxDiff;
		_wordMaps=new HashMap[_sequences.length];
	}
	
	public void process()
	{
		_wordLength=_minLength/(_maxDiff+1);
		if(_wordLength<MIN_WORD_LENGTH)
		{
			_startPositions[current]=_startPositions[previous];
			return;
		}
		
		hashWordsForFirstSequence();	
		
		for(int i=1;i<_sequences.length;i++)
		{
			hashWords(i);
		}
		
		for(int i=1;i<_sequences.length;i++)
		{
			Iterator iter=_wordMaps[0].keySet().iterator();
			while(iter.hasNext())
			{
				String currWord=(String)iter.next();
				List positions=(List)_wordMaps[0].get(currWord);
				if(_wordMaps[i].containsKey(currWord))
				{
					for(int j=0;j<positions.size();j++)
					{
						int start=((Integer)positions.get(j)).intValue();
						markPositions(start);
					}
				}
			}
			flipCurrentPrevious();
			_startPositions[current]=new BitSet();
			//System.out.println(Util.bitSetToString(_startPositions[previous],_sequences[0].length));
		}
		flipCurrentPrevious();
	}
	
	
	private void markPositions(int start)
	{
		int from=Math.max(0,start-(_minLength-_wordLength));
		int to=start;
		for(int i=from;i<=to;i++)
		{
			if(_startPositions[previous].get(i))
				_startPositions[current].set(i);
		}
	}
	
	private void flipCurrentPrevious()
	{
		if(current==0)
		{
			current=1;
			previous=0;
		}
		else
		{
			current=0;
			previous=1;
		}
	}
	
	private  void hashWords(int seqIndex)
	{
		String currSeq=new String(_sequences[seqIndex]);
		_wordMaps[seqIndex]=new HashMap();
		for(int i=0;i<(currSeq.length()-_wordLength+1);i++)
		{
			String word=currSeq.substring(i,i+_wordLength);
			_wordMaps[seqIndex].put(word,null);
		}
	}
	
	private void hashWordsForFirstSequence()
	{
		String currSeq=new String(_sequences[0]);
		_wordMaps[0]=new HashMap();
		
		for(int i=0;i<(currSeq.length()-_wordLength+1);i++)
		{
			String word=currSeq.substring(i,i+_wordLength);
			List positionsForWord=(List)_wordMaps[0].get(word);
			if(positionsForWord==null)
				positionsForWord=new LinkedList();
			positionsForWord.add(new Integer(i));
			_wordMaps[0].put(word,positionsForWord)	;		
		}
	}
	
	public BitSet getFinalStartPositions()
	{
		return _startPositions[current];
	}
}
