package ca.virology.japafi.application;
import java.util.*;

import ca.virology.japafi.algorithms.*;
import ca.virology.japafi.utils.*;
import ca.virology.japafi.structures.*;
/**
 * Finds in a set of input sequences groups of similar local substrings
 * Input: set of sequences, minimum length of pattern, max number of allowed errors
 * Output: Groups of patterns in a format:
 * Consensus substring
 * substrings from each input sequence 
 * wich have up to maxDiff errors with consensus substring.
 * (Note, that between them all patterns in the group may have up to 2*maxDiff errors)
 * @author marina
 *
 */
public class PatternFinder 
{
	char [][] _sequences;
	int _minLength;
	int _maxDiff;
	BitSet [] _startPositions=new BitSet[2];
	private int current=1;
	private int previous=0;
	
	List _finalPatterns=new LinkedList();
	Map _consensusSubstrings=new HashMap();
	String _firstString;
	String [] _sequenceNames;
	public PatternFinder(char [][] sequences, int minLen, int maxDiff)
	{
		_sequences=sequences;
		_minLength=minLen;
		_maxDiff=maxDiff;
		_startPositions[previous]=new BitSet(_sequences[0].length);
		_startPositions[previous].set(0,_sequences[0].length-1);
		_firstString=new String(_sequences[0]);
	}
	
	public void preProcess()
	{
		//cheap filtering based on exact matches
		Filter filter=new Filter(_startPositions[previous],_sequences,_minLength, _maxDiff);
		filter.process();
		_startPositions[current]=filter.getFinalStartPositions();
		
		flipCurrentPrevious();
		for(int i=1;i<_sequences.length;i++)
		{
			char [] currSeq=_sequences[i];
			APBTBoolean apbtBool=new APBTBoolean(_startPositions[previous],_sequences[0],
					currSeq,_minLength,_maxDiff);
			apbtBool.process();
			_startPositions[current]=apbtBool.getNewStartPositions();
			flipCurrentPrevious();
		}
		
		flipCurrentPrevious();
		for(int i=0;i<_sequences[0].length;i++)
		{
			if(_startPositions[current].get(i))
			{
				String substr=_firstString.substring(i, Math.min(i+_minLength+_maxDiff,_firstString.length()));
				
				_consensusSubstrings.put(substr,null);
			}
		}
	}
	
	public int getTotalPatternGroups()
	{
		return _consensusSubstrings.size();
	}
	
	public void process()
	{
		Iterator iter=_consensusSubstrings.keySet().iterator();
		
		
		while(iter.hasNext())
		{
			String patternAsString=(String)iter.next();
			char [] pattern=patternAsString.toCharArray();
			
			PatternGroup group=new PatternGroup(_sequences.length);
			group.consensus=patternAsString;
			boolean isMissing=false;
			for(int j=0;j<_sequences.length && !isMissing;j++)
			{
				APBTPattern patternB=new APBTPattern(_sequences[j],pattern,_minLength, _maxDiff);
				patternB.process();
				List solutionsForsequenceJ=patternB.getSolutions();
				if(solutionsForsequenceJ.size()==0)
				{
					isMissing=true;
				}
				else
					group.positions[j]=solutionsForsequenceJ;
			}
			if(!isMissing)
				_finalPatterns.add(group);
		}
	}
	
	public List getSolutions()
	{
		return _finalPatterns;
	}
	
	public void setSequenceNames(String [] names)
	{
		_sequenceNames=names;
	}
	
	public StringBuffer getDelimitedResults()
	{
		StringBuffer ret=new StringBuffer("");
		String [] strings=new String[_sequences.length];
		for(int i=0;i<_sequences.length;i++)
			strings[i]=new String(_sequences[i]);
		
		for(int i=0;i<_finalPatterns.size();i++)
		{
			PatternGroup pgroup=(PatternGroup)_finalPatterns.get(i);
			ret.append("Pattern group "+(i+1)+": "+pgroup.consensus+System.getProperty("line.separator"));
			for(int j=0;j<pgroup.positions.length;j++)
			{
				String seqname=getSequenceName(j);
				List positions=pgroup.positions[j];
				for(int k=0;k<positions.size();k++)
				{
					Integer pos=(Integer)positions.get(k);
					String substring=strings[j].substring(pos.intValue(),Math.min(pos.intValue()+_minLength+_maxDiff,strings[j].length()));
					ret.append(seqname+"\t"+pos+"\t"+substring+System.getProperty("line.separator"));
				}
				
			}
			ret.append(System.getProperty("line.separator"));
		}
		
		
		return ret;
	}
	
	private String getSequenceName(int seqIndex)
	{
		if(_sequenceNames==null)
			return "Sequence "+(seqIndex+1);
		else
			return _sequenceNames[seqIndex];
	}
	public void printSolutions()
	{
		String [] strings=new String[_sequences.length];
		for(int i=0;i<_sequences.length;i++)
			strings[i]=new String(_sequences[i]);
		
		for(int i=0;i<_finalPatterns.size();i++)
		{
			PatternGroup pgroup=(PatternGroup)_finalPatterns.get(i);
			System.out.println("Pattern group "+(i+1)+": "+pgroup.consensus);
			for(int j=0;j<pgroup.positions.length;j++)
			{
				System.out.println("Occurences in sequence "+getSequenceName(j)+" :");
				List positions=pgroup.positions[j];
				for(int k=0;k<positions.size();k++)
				{
					Integer pos=(Integer)positions.get(k);
					String substring=strings[j].substring(pos.intValue(),Math.min(pos.intValue()+_minLength+_maxDiff,strings[j].length()));
					System.out.println(pos+": "+substring);
				}
			}
		}
	}
	
	private void flipCurrentPrevious()
	{
		if(current==0)
		{
			current=1;
			previous=0;
		}
		else
		{
			current=0;
			previous=1;
		}
	}
	
	public long estimatedPreprocessingTime()
	{
		int N=1000;
		char [] alphabet={'a','c','g','t'};
		char [] seq1_1000=RandomStringGenerator.getRandomString(N,alphabet).toCharArray();
		char [] seq2_1000=RandomStringGenerator.getRandomString(N,alphabet).toCharArray();
		
		BitSet initialStartPositions=new BitSet(1000);
		initialStartPositions.set(0,999);
		
		long start=System.currentTimeMillis();
		APBTBoolean apbtBool=new APBTBoolean(initialStartPositions,seq1_1000,
				seq2_1000,_minLength,_maxDiff);
		apbtBool.process();
		long howLong1000_1000=System.currentTimeMillis()-start;
		
		long totalestimatedTimeinsec=howLong1000_1000/1000;
		long len1=_sequences[0].length;
		long len2=_sequences[1].length;
			
		long estimatedTime=howLong1000_1000*(len1/1000*len2/1000);
		//rough estimation - 2 first sequences multiplied by number_of_sequences-2
		if(_sequences.length==2)
			return estimatedTime;
		return ((_sequences.length-2)*estimatedTime);
	}
	
	public long estimatedProcessingTime()
	{
		Iterator iter=_consensusSubstrings.keySet().iterator();
		String substr=(String)iter.next();
		
		long start=System.currentTimeMillis();
		
		APBTPattern patternB=new APBTPattern(_sequences[0],substr.toCharArray(),_minLength, _maxDiff);
		patternB.process();
		long howLongOnePattern=System.currentTimeMillis()-start;
		
		return howLongOnePattern*_consensusSubstrings.size()*_sequences.length;
	}
}
