package ca.virology.japafi.application;

import java.io.*;

import java.util.BitSet;

import ca.virology.japafi.utils.SequenceFileReader;

public class TextApplication {
	
	public static final int MAXIMUM_PATTERN_GROUPS=1000;
	public static final int MINIMUM_LENGTH_TO_ESTIMATE_RUNNING_TIME=20000;
	
	public static void main(String [] args)
	{
		String [] files=null ;		
		
		int maxDiff=0;
		int minLen=0;			
		
		try
		{		
			minLen=Integer.parseInt(args[0]);
			maxDiff=Integer.parseInt(args[1]);
			
			files=new String [args.length-2] ;	
			for(int i=2;i<args.length;i++)
			{
				files[i-2]=args[i];
			}			
		}
		catch(Exception e)
		{
			System.out.println("Usage: \n" + 
					"java -Xmx512M -Xms512m -classpath classes ca.virology.japafi.algorithms.TextApplication \\ \n" +
					"<minLength> <maxDifferences> \n"+
					"<filename1> <filename2> ... <filenameN>" 
            		 );
			System.exit(1);
		}

		if(maxDiff==0)
		{
			System.out.println("This program is designed for APPROXIMATE common patterns");
			System.out.println("To find EXACT matches use the Suffix tree based application in VGO.");
			System.exit(0);
		}
		
		System.out.println("Java Pattern Finder");					
		System.out.println("***********************************************");
		System.out.println();
		
		char [][] sequences=new char [files.length][];
		for(int i=0;i<sequences.length;i++)
		{
			SequenceFileReader reader=new SequenceFileReader(files[i]);
			sequences[i]=reader.getSequence().toCharArray();
		}
		
		BitSet oldStartPositions=new BitSet(sequences[0].length);
		oldStartPositions.set(0,sequences[0].length-1);
		
		long start=System.currentTimeMillis();
		
		PatternFinder finder=new PatternFinder(sequences,minLen, maxDiff);
		
		if(sequences[0].length>MINIMUM_LENGTH_TO_ESTIMATE_RUNNING_TIME)
		{
			System.out.println("Calculating approximate running time. Please wait...");
			long estimatedPreprocessTime=finder.estimatedPreprocessingTime();
			System.out.println("To estimate the size of an output will take APPROXIMATELY: "+estimatedPreprocessTime/1000+" seconds.");
			System.out.println("Thank you for your patience.");	System.out.println("***********************************************");
			System.out.println();
			
		}
		
		finder.setSequenceNames(files);
		System.out.println("Now calculating total number of common patterns. Please wait....");
		finder.preProcess();
		System.out.println("Processed in "+((System.currentTimeMillis()-start)/1000)+" seconds.");
		int totalGroups=finder.getTotalPatternGroups();
		
		if(totalGroups==0)
		{
			System.out.println("There are no matching substrings of length "+minLen+" with up to "+maxDiff+" differences in the input sequences.");
			System.exit(0);
		}
		
		System.out.println("Maximum number of pattern groups : "+totalGroups);
		System.out.println("***********************************************");
		System.out.println();
		
		if(totalGroups>MAXIMUM_PATTERN_GROUPS)
		{
			System.out.println("There are too many matching substrings of length "+minLen+" with up to "+maxDiff+" differences in the input sequences.");
			System.out.println("It is recommanded that you \n either decrease the number of allowed differences, \n or increase the minimum pattern length.");
			System.out.println("Then run Java Pattern Finder with new parameters.\n");
			System.out.println("***********************************************");
			System.out.println();
			
			
			System.out.println("Do you still want to collect the start positions of all "+totalGroups+" patterns in "+sequences.length+" input sequences? y/n");
			
			try
			{
				BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
				String yesno = br.readLine();
				if(!yesno.toLowerCase().startsWith("y"))
				{
					System.exit(1);
				}
				System.out.println("***********************************************");
				System.out.println();
				
				System.out.println("Calculating approximate running time. Please wait...");
				
				long estimatedProcessTime=finder.estimatedProcessingTime();
				System.out.println("To find  positions of "+totalGroups+" patterns will take APPROXIMATELY: "+estimatedProcessTime/1000+" seconds.");
				System.out.println("Do you still want to continue? y/n");
				br = new BufferedReader(new InputStreamReader(System.in));
				yesno = br.readLine();
				if(!yesno.toLowerCase().startsWith("y"))
				{
					System.exit(1);
				}		
				System.out.println("***********************************************");
				System.out.println();
				
			}
			catch(Exception e)
			{
				System.exit(1);
			}
		}
		
		System.out.println("Locating positions of "+totalGroups+" common patterns in "+sequences.length+" sequences. Please wait...");
		
		
		start=System.currentTimeMillis();
		finder.process();
		
		if(finder.getSolutions().size()==0)
		{
			System.out.println("After refined search there are no matching substrings of length "+minLen+" with up to "+maxDiff+" differences in the input sequences.");
			System.exit(0);
		}
		
		System.out.println("Collected all starting positions in "+((System.currentTimeMillis()-start)/1000)+" seconds.");
		System.out.println("***********************************************");
		System.out.println();
		
		System.out.println("Print patterns to the console? y/n");
		try
		{
			BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
			String yesno = br.readLine();
			if(yesno.toLowerCase().startsWith("y"))
			{
				finder.printSolutions();
			}
		}
		catch(Exception e)
		{
			System.exit(1);
		}
		
		System.out.println("***********************************************");
		System.out.println();
		
		System.out.println("Save patterns to file? y/n");
		try
		{
			BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
			String yesno = br.readLine();
			if(!yesno.toLowerCase().startsWith("y"))
			{
				System.exit(1);
			}
			else
			{
				System.out.println("Enter the name of the file you want to save the results to:");
				br = new BufferedReader(new InputStreamReader(System.in));
				String filename = br.readLine();
				StringBuffer sbresult=finder.getDelimitedResults(); 
				FileWriter fstream = new FileWriter(filename);
		        BufferedWriter out = new BufferedWriter(fstream);
		        out.write(sbresult.toString());
		    
		        out.close();
			}
		}
		catch(Exception e)
		{
			System.out.println("Unexpected error: "+e.getMessage());
			System.exit(1);
		}
		
		System.out.println("********THE END************");
		
		
	}
}
